require 'pygments.rb'
require 'json'

module Roxie
  module CLI

    class Doc < Thor::Group

      include Thor::Actions

      class << self
        def banner
          "#{basename} doc [INPUT]"
        end
      end

      namespace :doc

      check_unknown_options!
      add_runtime_options!

      argument 'paths',
        :type     => :array,
        :default  => ['{lib,app}/**/*.rb', 'ext/**/*.c']

      class_option 'base-path',
        :aliases  => '-p',
        :default  => './',
        :desc     => "The path to look for the input files in"
      class_option 'output',
        :aliases  => '-o',
        :default  => './doc/api.json',
        :desc     => "The output file (JSON)"
      class_option 'exclude',
        :aliases  => '-x',
        :type     => :array,
        :desc     => "Directories to ignore (regexp)"
      class_option 'language',
        :aliases  => '-l',
        :desc     => "Language of input files, if it can't be inferred"
      class_option 'only-tagged',
        :type     => :boolean,
        :default  => false,
        :desc     => "Only extract comments that have @tags in them"
      class_option 'only-single',
        :type     => :boolean,
        :default  => false,
        :desc     => "Only extract single-line comments"
      class_option 'only-multi',
        :type     => :boolean,
        :default  => false,
        :desc     => "Only extract multi-line comments"

      def parse_paths
        excluded = (options[:exclude] || []).map do |path|
          case path
          when Regexp; path
          else Regexp.new(path.to_s, Regexp::IGNORECASE)
          end
        end
        inside options[:'base-path'] do
          @files = paths.
            map {|p| File.directory?(p) ? "#{p}/**/*" : p }.
            map {|p| p.include?("*") ? Dir[p].sort_by {|f| f.length } : p }.flatten.
            reject {|p| !File.file?(p) || excluded.any? {|re| p =~ re } }
        end
      end

      def parse_in_order
        @docs = {}

        while file = @files.shift
          language =
            if detected_lang = Pygments::Lexer.find_by_extname(File.extname(file))
              detected_lang[:name]
            else
              options[:language]
            end
          data = File.read(File.join(options['base-path'], file))

          opts = {
            :language => language,
            :only_tagged => options[:'only-tagged'],
            :only_single => options[:'only-single'],
            :only_multi => options[:'only-multi']
          }.reject { |k, v| v.nil? }

          # Extract relevant comments out of file
          if doc = Roxie.extract(data, opts)
            @docs[file] = doc unless doc.empty?
          end
        end
      end

      def generate_output!
        create_file options[:output] do
          JSON.pretty_generate @docs
        end
      end

    end

  end
end
