require 'strscan'

module Roxie

  # Extract comments and tags from text
  class Extractor

    def self.default_options
      {
        :language      => 'Ruby',
        :only_tagged   => false,
        :only_single   => false,
        :only_multi    => false
      }
    end

    def initialize(options = {})
      @options = self.class.default_options.merge(options)
      @language = LANGUAGES[@options[:language]]
    end

    # Read up to 100KB
    BYTE_LIMIT = 100_000

    def extract(text)
      return unless @language

      @code = text.dup
      @scanner = StringScanner.new(@code)

      # Extract comments from text
      extract_comments
    end

  private

    TAG = /^@(\w+)[ \t]*(.*$)/.freeze

    def matchers
      @_matchers ||= @language[:matchers]
    end

    def extract_comments
      comments = []
      until @scanner.eos?
        break if @scanner.pos >= BYTE_LIMIT

        # Skip over shebangs and encoding information
        if @scanner.scan(/^#!.+$/) || @scanner.scan(/^#.*coding[=:]\s*[-\w.]+.*$/)
          # just consume it
        elsif comment_block =
          # Single-line comment
          if !@options[:only_multi] && matchers[:start_single] &&
            @scanner.bol? && @scanner.scan(matchers[:start_single])
            extract_single
          # Multi-line comment
          elsif !@options[:only_single] && matchers[:start_multi] &&
            @scanner.scan(matchers[:start_multi])
            extract_multi
          end

          # Now that we have our block, let's strip out our tags and
          # add it to our list
          tags = extract_tags!(comment_block)
          comment = {
            :description => comment_block.join("\n").strip,
            :tags => tags
          }

          # Scan until the next non-whitespace character to get our line #
          @scanner.scan_until(/(?=\S)/)
          comment[:line] = @code[0..@scanner.pos].count("\n") + 1 unless @scanner.eos?

          comments << comment unless @options[:only_tagged] && tags.empty?
        else
          # Skip line
          @scanner.skip_until(/\n|\Z/)
        end
      end

      comments
    end

    def extract_single
      block = []
      loop do
        block << @scanner.scan_until(/\n|\Z/).rstrip
        break unless @scanner.scan(matchers[:start_single])
      end
      block
    end

    def extract_multi
      block = @scanner.scan_until(matchers[:end_multi_with])
      @scanner.skip_until(matchers[:end_multi])

      block.split("\n").map do |line|
        matchers[:middle_multi] ? line.sub(matchers[:middle_multi], '') : line
      end
    end

    def extract_tags!(block)
      tags = []
      block.reverse_each do |line|
        if match = line.match(TAG)
          tag = {
            :type => match[1],
            :description => match[2].strip
          }
          tags << tag

          # Remove this line from the comment block if we match
          block.delete line
        end
      end

      tags.reverse
    end

  end

end
